/*global fwRegisterOnInit, fwRegisterOnMouseUp, fwRegisterOnDraw, fwRegisterOnDrawHelp, fwRegisterGame*/
/*global font, backgroundColor, puzzleImage*/
/*global canvas, ctx, bgctx, addctx, draw*/
/*global fwFillRect, fwStrokeFillRect, fwFillText, fwPointInBox, fwDrawGameLevel*/
/*global noAutoNextPuzzle:true, hideGameBtn:true, fwRegisterOverlayCtx, fwAddImageToLoad*/
/*global fwRegisterOnLevelButtonUp,  fwRegisterOnOptionsSync, fwRegisterOnOptionsChanged, fwTriggerNextPicture*/
/*global fwPickLanguageText*/
/*global fwStartOptionsBox, optionsBox, gameLevel:true, gameLevels: true, getFwString, sidFw, startHighlightNext*/
/*global console*/
"use strict";
var g_restyled = g_restyled || undefined;
var drawBackground;
var digipuzzleGreen = 'rgb(0, 255, 33)';
var xImage = null;
var oImage = null;
var xImageFileName = xImageFileName || undefined;
var oImageFileName = oImageFileName || undefined;
var m_gridRedImage, m_gridGreenImage, m_gridWhiteImage, m_backgroundImage;
var gridColor = gridColor || undefined;
var m_gameLevelRow;
var m_nrGames = 0;

var PLANE_W = 180;
var PLANE_H = 180;
var PLANE_START_X = Math.floor((788 - (3 * PLANE_W)) / 2);
var PLANE_START_Y = Math.floor((606 - (3 * PLANE_H)) / 2);

var OPPONENT_ENUM = {
    COMPUTER:               0,
    FRIEND:                 1
};
var m_opponents = [OPPONENT_ENUM.COMPUTER, OPPONENT_ENUM.FRIEND];
var m_opponent = OPPONENT_ENUM.COMPUTER;
var m_opponentRow;

var sid = {
    SELECT_LEVEL:               0,
    CHOOSE_YOUR_OPPONENT:       1,
    COMPUTER:                   2,
    FRIEND:                     3,
};

var m_audioSpritesSample;

function getString(id) {
    var texts = [];
    switch (id) {
    case sid.SELECT_LEVEL:
        texts = ["Select a level",
            "Selecteer een niveau",
            "Wähle einen Schwierigkeitsgrad",
            "Vælg sværhedsgrad",
            "Selecciona el nivel",
            "Selecione o nível",
            "Erfiðleikastig",
            "Välj en nivå",
            "Velg et nivå",
            "Izberi stopnjo",
            "Sélectionner un niveau"];
        break;
    case sid.CHOOSE_YOUR_OPPONENT:
        texts = ["Choose your opponent",
            "Kies je tegenstander",
            "Wähle deinen Gegner",
            "Vælg din modstander",
            "Elige tu oponente",
            "Escolha seu adversário",
            "",
            "Välj din motståndare",
            "Velg motstanderen din",
            "Izberi nasprotnika",
            "Choisis ton adversaire"];
        break;
    case sid.COMPUTER:
        texts = ["Computer",
            "Computer",
            "Computer",
            "Computer",
            "Ordenador",
            "Computador",
            "",
            "Dator",
            "Datamaskin",
            "Računalnik",
            "Ordinateur"];
        break;
    case sid.FRIEND:
        texts = ["Friend",
            "Vriend",
            "Freund",
            "Ven",
            "Amigo",
            "Amigo",
            "",
            "Vän",
            "Venn",
            "Prijatelj",
            "Ami"];
        break;
    default:
        break;
    }
    return (fwPickLanguageText(texts));
}

var TURN_USER = 1;
var TURN_AI = -1;
var TURN_USER_2 = 2;
var turn = 0;
var xWon = 0;
var oWon = 0;
var catsGame = 0;
var cells = [];
var m_finished = false;

function makeCells() {
    var i;
    cells = [];
    for (i = 0; i < 9; i += 1) {
        cells.push({value: '', x: PLANE_START_X + ((i % 3) * PLANE_W), y: PLANE_START_Y + (Math.floor(i / 3) * PLANE_H), w: PLANE_W, h: PLANE_H, winner: false});
    }
}

function nextTurn() {
    if (turn === TURN_AI) {
        switch (gameLevel) {
        case 'easy':
            beginnerMove();
            break;
        case 'normal':
            intermediateMove();
            break;
        case 'hard':
            experiencedMove();
            break;
        case 'extreme':
            perfectMove();
            break;
        }
        turn  = TURN_USER;

        if (!m_finished && m_audioSpritesSample) {
            m_audioSpritesSample.play('o');
        }
        draw();
    } else {
        // wait for user move
/*    
        switch (gameLevel) {
        case 'easy': 
            beginnerMove();
            break;
        case 'normal': 
            intermediateMove();
            break;
        case 'hard': 
            experiencedMove();
            break;
        case 'extreme': 
            perfectMove();
            break;
        }
*/
    }
}

function getLegalMoves(state) {
    var i, moves = 0;
    for (i = 0; i < 9; i += 1) {
        if ((state & (1 << (i * 2 + 1))) === 0) {
            moves |= (1 << i);
        }
    }
    return moves;
}

function moveRandom(moves) {
    var i, j, numMoves = 0, moveNum;
    for (i = 0; i < 9; i += 1) {
        if ((moves & (1 << i)) !== 0) { numMoves += 1; }
    }
    if (numMoves > 0) {
        moveNum = Math.ceil(Math.random() * numMoves);
        numMoves = 0;
        for (j = 0; j < 9; j += 1) {
            if ((moves & (1 << j)) !== 0) { numMoves += 1; }
            if (numMoves === moveNum) {
                move(cells[j]);
                return;
            }
        }
    }
}

function openingBook(state) {
    var mask = state & 0x2AAAA;
    if (mask === 0x00000) { return 0x1FF; }
    if (mask === 0x00200) { return 0x145; }
    if ((mask === 0x00002) || (mask === 0x00020) || (mask === 0x02000) || mask === (0x20000)) { return 0x010; }
    if (mask === 0x00008) { return 0x095; }
    if (mask === 0x00080) { return 0x071; }
    if (mask === 0x00800) { return 0x11C; }
    if (mask === 0x08000) { return 0x152; }
    return 0;
}

function perfectMove() {
    var i, value, state = getState(), winner = detectWin(state), moves, hope, goodMoves;
    if (winner === 0) {
        moves = getLegalMoves(state);
        hope = -999;
        goodMoves = openingBook(state);
        if (goodMoves === 0) {
            for (i = 0; i < 9; i += 1) {
                if ((moves & (1 << i)) !== 0) {
                    value = moveValue(state, i, turn, turn, 15, 1);
                    if (value > hope) {
                        hope = value;
                        goodMoves = 0;
                    }
                    if (hope === value) {
                        goodMoves |= (1 << i);
                    }
                }
            }
        }
        moveRandom(goodMoves);
    }
}

function moveValue(istate, move, moveFor, nextTurn, limit, depth) {
    var i, state = stateMove(istate, move, nextTurn),
        winner = detectWin(state), moves, hope, value;
    if ((winner & 0x300000) === 0x300000) {
        return 0;
    } else if (winner !== 0) {
        if (moveFor === nextTurn) return 10 - depth;
        else return depth - 10;
    }
    hope = 999;
    if (moveFor !== nextTurn) { hope = -999; }
    if (depth === limit) { return hope; }
    moves = getLegalMoves(state);
    for (i = 0; i < 9; i += 1) {
        if ((moves & (1 << i)) !== 0) {
            value = moveValue(state, i, moveFor, -nextTurn, 10-Math.abs(hope), depth+1);
            if (Math.abs(value) !== 999) {
                if (moveFor === nextTurn && value < hope) {
                    hope = value;
                } else if (moveFor !== nextTurn && value > hope) {
                    hope = value;
                }
            }
        }
    }
    return hope;
}

function detectWinMove(state, cellNum, nextTurn) {
    var value = 0x3;
    if (nextTurn === -1) value = 0x2;
    var newState = state | (value << cellNum * 2);
    return detectWin(newState);
}

function beginnerMove() {
    var state = getState(),
        winner = detectWin(state);
    if (winner === 0) { moveRandom(getLegalMoves(state)); }
}

function getGoodMove(state){
    var i, j, moves = getLegalMoves(state);
    for (i = 0; i < 9; i += 1) {
        if ((moves & (1 << i)) !== 0) {
            if (detectWinMove(state, i, turn)) {
                move(cells[i]);
                return 0;
            }
        }
    }
    for (j = 0; j < 9; j += 1){
        if ((moves & (1 << j)) !== 0) {
            if (detectWinMove(state, j, -turn)) {
                move(cells[j]);
                return 0;
            }
        }
    }
    return moves;
}

function intermediateMove() {
    var state = getState();
    var winner = detectWin(state);
    if (winner === 0) {
        moveRandom(getGoodMove(state));
    }
}

function experiencedMove() {
    var state = getState();
    var winner = detectWin(state);
    if (winner === 0) {
        var moves = openingBook(state);
        if (state === 0) moves = 0x145;
        if (moves === 0) moves = getGoodMove(state);
        moveRandom(moves);
    }
}

function getState() {
    var i, state = 0;
    for (i = 0; i < 9; i += 1) {
        var cell = cells[i];
        var value = 0;
        if (cell.value.indexOf('X') != -1) value = 0x3;
        if (cell.value.indexOf('O') != -1) value = 0x2;
        state |= value << (i * 2);
    }
    return state;
}

function detectWin(state) {
    if ((state & 0x3F000) === 0x3F000) { return 0x13F000; }
    if ((state & 0x3F000) === 0x2A000) { return 0x22A000; }
    if ((state & 0x00FC0) === 0x00FC0) { return 0x100FC0; }
    if ((state & 0x00FC0) === 0x00A80) { return 0x200A80; }
    if ((state & 0x0003F) === 0x0003F) { return 0x10003F; }
    if ((state & 0x0003F) === 0x0002A) { return 0x20002A; }
    if ((state & 0x030C3) === 0x030C3) { return 0x1030C3; }
    if ((state & 0x030C3) === 0x02082) { return 0x202082; }
    if ((state & 0x0C30C) === 0x0C30C) { return 0x10C30C; }
    if ((state & 0x0C30C) === 0x08208) { return 0x208208; }
    if ((state & 0x30C30) === 0x30C30) { return 0x130C30; }
    if ((state & 0x30C30) === 0x20820) { return 0x220820; }
    if ((state & 0x03330) === 0x03330) { return 0x103330; }
    if ((state & 0x03330) === 0x02220) { return 0x202220; }
    if ((state & 0x30303) === 0x30303) { return 0x130303; }
    if ((state & 0x30303) === 0x20202) { return 0x220202; }
    if ((state & 0x2AAAA) === 0x2AAAA) { return 0x300000; }
    return 0;
}
/*
function recordWin(winner) {
    "use strict";
    if ((winner & 0x300000) === 0x100000) {
        xWon++;
    } else if ((winner & 0x300000) === 0x200000) {
        oWon++;
    } else if ((winner & 0x300000) === 0x300000) {
        catsGame++;
    }
    drawStats();
}*/

function drawStats() {
/*    
    var b = document.board;
    var totalGames = xWon + oWon + catsGame;
    b.xWon.value = xWon;
    b.oWon.value = oWon;
    b.catsGame.value = catsGame;
    b.xWonPer.value = ((xWon==0)?0:(Math.round(xWon * 1000 / totalGames) / 10)) + '%';
    b.oWonPer.value = ((oWon==0)?0:(Math.round(oWon * 1000 / totalGames) / 10)) + '%';
    b.catsGamePer.value = ((catsGame==0)?0:(Math.round(catsGame * 1000 / totalGames) / 10)) + '%';
    var cookie = ':' + xWon + ':' + oWon + ':' + catsGame;
    var expires = new Date();
    // cookie expires in one year
    expires.setTime(expires.getTime() + 365 * 24 * 60 * 60 * 1000);
    document.cookie = (
    'tictactoeStats' + '=' +
    escape(cookie) +
    '; expires=' + expires.toGMTString()
    );*/
}
/*
function clearStats() {
    xWon = 0;
    oWon = 0;
    catsGame = 0;
    drawStats();
}
*/
function drawState(state) {
    var i, winner = detectWin(state);
    if ((winner & 0x300000) !== 0) {
        if ((winner & 0x300000) === 0x100000) {
            if (!m_finished && m_audioSpritesSample) {
                m_audioSpritesSample.play('win');
            }
            xWon++;
        } else if ((winner & 0x300000) === 0x200000) {
            if (!m_finished && m_audioSpritesSample) {
                if (m_opponent === OPPONENT_ENUM.FRIEND) {
                    m_audioSpritesSample.play('win');
                } else {
                    m_audioSpritesSample.play('lose');
                }
            }
            oWon++;
        } else {
            if (!m_finished && m_audioSpritesSample) {
                m_audioSpritesSample.play('tie');
            }
            catsGame++;
        }
        if (!m_finished) {
            m_finished = true;
            startHighlightNext();
        }
        drawStats();
    }
    for (i = 0; i < 9; i += 1) {
        var value = '';
        if ((state & (1 << (i * 2 + 1))) !== 0) {
            if ((state & (1 << (i * 2))) !== 0) {
                value = 'X';
            } else {
                value = 'O';
            }
        }
        if ((winner & (1 << (i * 2 + 1))) !== 0) {
            cells[i].winner = true;
        } else {
            cells[i].winner = false;
        }
        cells[i].value = value;
    }
}

function stateMove(state, move, nextTurn) {
    var value = 0x3;
    if (nextTurn == -1) value = 0x2;
    return (state | (value << (move*2)));
}

function move(cell){
    var i;
    if (cell.value === '') {
        var state = getState();
        var winner = detectWin(state);
        if (winner === 0){
            for (i = 0; i < 9; i += 1) {
                if (cells[i] === cell) {
                    state = stateMove(state, i, turn);
                }
            }
            drawState(state);
        }
    }
}

function countMoves(state) {
    var i;
    var count = 0;
    for (i = 0; i < 9; i += 1) {
        if ((state & (1 << (i * 2 + 1))) !== 0){
            count++;
        }
    }
    return count;
}

function newGame() {
    makeCells();
    var state = getState();
    var winner = detectWin(state);
    if (winner === 0 && countMoves(state) > 1) {
        if (turn === 1) oWon++;
        else xWon++;
        drawStats();
    }
    drawState(0);
    
    if ((m_nrGames % 2) === 0) {
        turn = TURN_USER;
    } else {
        if (m_opponent === OPPONENT_ENUM.COMPUTER) {
            turn = TURN_AI;
            setTimeout(nextTurn, 500);
        } else {
            turn = TURN_AI;
        }
    }
    m_nrGames += 1;
    m_finished = false;
}

function ticTacToeOnDraw() {
    var i, cell, image, x, w;
    if (drawBackground) {
        if (m_backgroundImage) {
            bgctx.drawImage(m_backgroundImage, 0, 0);
            fwFillRect(bgctx, canvas.width - 84, 0, 84, canvas.height, 'rgba(255,255,255,0.25)');
            fwFillRect(bgctx, PLANE_START_X - 1, PLANE_START_Y - 1, (3 * PLANE_W) + 2,  (3 * PLANE_H) + 2, '#4C4A8D');
        } else {
            if (puzzleImage.width === 862) {
                bgctx.drawImage(puzzleImage, 0, 0);
            } else {
                bgctx.fillStyle = backgroundColor;
                bgctx.fillRect(0, 0, canvas.width, canvas.height);
                bgctx.drawImage(puzzleImage, 0, 0, puzzleImage.width, puzzleImage.height, 5, 5, canvas.width - 10, canvas.height - 10);
            }
        }
        drawBackground = false;
    }
    ctx.clearRect(0, 0, canvas.width, canvas.height);
    
    for (i = 0; i < cells.length; i += 1) {
        cell = cells[i];
        if (g_restyled === true) {
            if (cell.winner) {
                image = (cell.value === 'X') ? m_gridGreenImage : m_gridRedImage;
            } else {
                image = m_gridWhiteImage;
            }
            console.log('cell.x: ' + cell.x + ', cell.y: ' + cell.y);
            ctx.drawImage(image, cell.x, cell.y);
            if (cell.value === 'X') {
                ctx.drawImage(xImage, cell.x, cell.y);
            } else if (cell.value === 'O') {
                ctx.drawImage(oImage, cell.x, cell.y);
            }
        } else {
            fwStrokeFillRect(ctx, cells[i].x, cells[i].y, cells[i].w, cells[i].h, 'black', 3, cells[i].winner ? ((cells[i].value === 'X') ? digipuzzleGreen : 'red') : (gridColor !== undefined) ? gridColor : 'white');
            if ((cells[i].value === 'X') && (xImage !== null)) {
                ctx.drawImage(xImage, cells[i].x, cells[i].y);
            } else if ((cells[i].value === 'O') && (oImage !== null)) {
                    ctx.drawImage(oImage, cells[i].x, cells[i].y);
            } else {
                fwFillText(ctx, cells[i].value, cells[i].x + (cells[i].w / 2), cells[i].y + (cells[i].h / 2) + 56, '112pt ' + font, 'black', 'center');
            }
        }
    }

    if (g_restyled === true) {
        w = 90;

        x = 52;
        ctx.save();
        ctx.globalAlpha = (!m_finished && (turn === TURN_USER)) ? 1 : 0.5;
        ctx.drawImage(xImage, 0, 0, xImage.width, xImage.height, x - Math.floor(w / 2), 10, w, w);
        ctx.restore();
        fwFillText(ctx, xWon.toString(), x + 2, 160 + 2 + 24, '48pt ' + font, '#F91E04', 'center');
        fwFillText(ctx, xWon.toString(), x, 160 + 24, '48pt ' + font, 'white', 'center');

        x = 778 - 52;
        ctx.save();
        ctx.globalAlpha = (!m_finished && (turn !== TURN_USER)) ? 1 : 0.5;
        ctx.drawImage(oImage, 0, 0, oImage.width, oImage.height, x - Math.floor(w / 2), 10, w, w);
        ctx.restore();
        fwFillText(ctx, oWon.toString(), x + 2, 160 + 2 + 24, '48pt ' + font, '#5DD410', 'center');
        fwFillText(ctx, oWon.toString(), x, 160 + 24, '48pt ' + font, 'white', 'center');
    }

    //fwDrawGameLevel(ctx);
}

// Method called after framework loaded all images
function ticTacToeOnInit() {
    drawBackground = true;
    newGame();
}

function ticTacToeOnMouseUp(mouseX, mouseY) {
    var i;
    if (!m_finished) {
        if (turn === TURN_USER) {
            for (i = 0; i < cells.length; i += 1) {
                if ((cells[i].value === '') && fwPointInBox(mouseX, mouseY, cells[i].x, cells[i].y, cells[i].w, cells[i].h)) {
                    move(cells[i]);
                    if (!m_finished && m_audioSpritesSample) {
                        m_audioSpritesSample.play('x');
                    }
                    if (m_opponent === OPPONENT_ENUM.COMPUTER) {
                        turn = TURN_AI;
                        setTimeout(nextTurn, 500);
                    } else {
                        turn = TURN_AI;
                    }
                    draw();
                }
            }
        } else if (m_opponent === OPPONENT_ENUM.FRIEND) {
            for (i = 0; i < cells.length; i += 1) {
                if ((cells[i].value === '') && fwPointInBox(mouseX, mouseY, cells[i].x, cells[i].y, cells[i].w, cells[i].h)) {
                    move(cells[i]);
                    if (!m_finished && m_audioSpritesSample) {
                        m_audioSpritesSample.play('o');
                    }
                    turn = TURN_USER;
                    draw();
                }
            }            
        }
    }
}

function ticTacToeOnDrawHelp() {
}

// called before options box is opened
function ticTacToeOnOptionsSync() {
    var i;
    if (m_gameLevelRow) {
        for (i = 0; i < m_gameLevelRow.items.length; i += 1) {
            m_gameLevelRow.items[i].selected = (gameLevel === gameLevels[i]);
        }
    }
    if (m_opponentRow) {
        for (i = 0; i < m_opponentRow.items.length; i += 1) {
            m_opponentRow.items[i].selected = (m_opponent === m_opponents[i]);
        }
    }
}

// called when options box is closed (with ok button)
function ticTacToeOnOptionsChanged() {
    var i, changed = false;
    if (m_gameLevelRow) {
        for (i = 0; i < m_gameLevelRow.items.length; i += 1) {
            if (m_gameLevelRow.items[i].selected !== (gameLevel === gameLevels[i])) {
                if (m_gameLevelRow.items[i].selected) {
                    gameLevel = gameLevels[i];
                }
                changed = true;
            }
        }
    }
    if (m_opponentRow) {
        for (i = 0; i < m_opponentRow.items.length; i += 1) {
            if (m_opponentRow.items[i].selected !== (m_opponent === m_opponents[i])) {
                if (m_opponentRow.items[i].selected) {
                    m_opponent = m_opponents[i];
                }
                changed = true;
            }
        }
    }
    if (changed) {
        m_nrGames = oWon = xWon = 0;
        fwTriggerNextPicture(1);
    }
}

function ticTacToeOnLevelButtonUp() {
    fwStartOptionsBox();
    draw();
}

function ticTacToeOnConfigure() {
    var i, items;

    hideGameBtn = true;
    noAutoNextPuzzle = true;
    gameLevels = ['easy', 'normal', 'hard'/*, 'extreme'*/];

    items = [];
    for (i = 0; i < gameLevels.length; i += 1) {
        items.push({selected: false, text: getFwString(sidFw.LEVEL_EASY + i)});
    }
    m_gameLevelRow = {type: 'bar', items: items, text: getString(sid.SELECT_LEVEL)};
    optionsBox.rows.push(m_gameLevelRow);

    items = [{selected: false, text: getString(sid.COMPUTER)}, {selected: false, text: getString(sid.FRIEND)}];
    m_opponentRow = {type: 'bar', items: items, text: getString(sid.CHOOSE_YOUR_OPPONENT)};
    optionsBox.rows.push(m_opponentRow);

    fwRegisterOnInit(ticTacToeOnInit);
    fwRegisterOnMouseUp(ticTacToeOnMouseUp);
    fwRegisterOnDraw(ticTacToeOnDraw);
    fwRegisterOnDrawHelp(ticTacToeOnDrawHelp);
    fwRegisterOnLevelButtonUp(ticTacToeOnLevelButtonUp);
    fwRegisterOnOptionsSync(ticTacToeOnOptionsSync);
    fwRegisterOnOptionsChanged(ticTacToeOnOptionsChanged);
    fwRegisterOverlayCtx(addctx);

    if (g_restyled) {
        m_gridWhiteImage = new Image();
        fwAddImageToLoad(m_gridWhiteImage, "grid_white.png");
        m_gridGreenImage = new Image();
        fwAddImageToLoad(m_gridGreenImage, "grid_green.png");
        m_gridRedImage = new Image();
        fwAddImageToLoad(m_gridRedImage, "grid_red.png");

        xImageFileName = "cross.png";
        oImageFileName = "circle.png";

        m_backgroundImage = new Image();
        fwAddImageToLoad(m_backgroundImage, "tictactoe_background.jpg");

        PLANE_W = PLANE_H = 190;
        PLANE_START_X = Math.floor((778 - (3 * PLANE_W)) / 2);
        PLANE_START_Y = Math.floor((606 - (3 * PLANE_H)) / 2);

        m_audioSpritesSample = new Howl({
          src: ["sounds/tictactoe.mp3"],
          sprite: {
            x: [0, 500],
            o: [2000, 500],
            win: [4000, 1500],
            lose: [6000, 1500],
            tie: [8000, 1500]
          }
        });
        audioEnabled = true;
    }

    if (xImageFileName !== undefined) {
        xImage = new Image();
        fwAddImageToLoad(xImage, xImageFileName);
    }
    if (oImageFileName !== undefined) {
        oImage = new Image();
        fwAddImageToLoad(oImage, oImageFileName);
    }
}

// register the game so it is (automatically) added to the menu
fwRegisterGame(ticTacToeOnConfigure);
